#include "disc_discords.h"
#include "utils.h"
#include "io_utils.h"
#include "globals.h"
#include <iostream>
#include <sstream>
#include <ctime>
#include <cmath>

using namespace std;

DiscDiscords::DiscDiscords(const char* dir, int k_discords, double threshold, 
						   int db_pages, int dist_type, int page_size, int ts_len) {
	this->dir = dir;
	this->threshold = threshold;
	this->db_pages = db_pages;
	this->ts_len = ts_len;
	this->k_disc = k_discords;
	this->page_size = page_size;
	this->dist_type = dist_type;

	string strdir = dir;
    LOG.open((strdir + "Output_discorddetector.txt").c_str());
	MEMLOG.open((strdir + "Output_memory.txt").c_str());
	DISCORDSLOG.open((strdir + "Output_discords.txt").c_str());
	DISCINDSLOG.open((strdir + "Output_discordinds_dists.txt").c_str());

	mainMem = new MainMemory(dir, threshold, db_pages, dist_type, page_size, ts_len);
	mainMem->LOG = &LOG;
	mainMem->MEMLOG = &MEMLOG;
}

DiscDiscords::~DiscDiscords() {
	delete mainMem;
	cleanContainer(discords);
	LOG.close();
	MEMLOG.close();
	DISCORDSLOG.close();
	DISCINDSLOG.close();
}

void DiscDiscords::findDiscords() {
	buildMemory();
	cout << "(Phase2) Finding the discords.............." << endl;

	int starttimer = clock();
    int offset = 0;
	vector <float*> DB;
	for (int i = 0; i < db_pages; i++){
        cout << "---(Phase2) Checking page#" << (i+1) << " of " << db_pages << endl;
        loadDB(DB, i);
		updateNNbs(DB, offset);
        offset += (int) DB.size();
		cleanContainer(DB);
		if (db_pages > 1 && i < (db_pages - 1)) {
			this->page_size = -1;
			this->ts_len = -1;
		}
		MEMLOG << (int) mainMem->MM.size() << " " << avr_comparisons[i] << endl;
	}
	cout << "End memmory size = " << (int) mainMem->MM.size() << endl;
	LOG << "End memmory size = " << (int) mainMem->MM.size() << endl;
	if (k_disc > (int) mainMem->MM.size()) {
		cout << "There are less elements in memory than required discords! Available memory elements " << (int) mainMem->MM.size() << endl;
		k_disc = (int) mainMem->MM.size();
	}
	k_disc = (int) mainMem->MM.size();
	selectTopDiscords();
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	cout << "(Phase2) Time to find discords: " << time_spent << "sec." << endl;
	LOG << "(Phase2) Time to find discords: " << time_spent << "sec." << endl;

	logDiscords();
}

void DiscDiscords::buildMemory() {
	mainMem->buildMemory();
	//if (this->page_size == -1) {
	//	page_size = mainMem->page_size;
	//	ts_len = mainMem->ts_len;
	//}
	for (int i = 0; i < (int) mainMem->MM.size(); i++) {
		nn_dists.push_back(10000000); 
		nnbs.push_back(-1);
	}
}

//void DiscDiscords::updateNNbs(vector<float*> &DB, int offset) {
//	double sqrThreshold = threshold*threshold;
//	int tmp_avr_comparisons = 0; //how many comparisons on average are necessary before discarding a DB element
//	for (int i = 0; i < (int) DB.size(); i++) {
//		float* dbEl = DB[i];
//		vector <float*>::iterator mm_Iter;
//		vector <int>::iterator indxPtrIter = mainMem->indxPtr.begin();
//		vector <float>::iterator nndistIter = nn_dists.begin();
//		vector <float>::iterator nnbsIter = nnbs.begin();
//		//cout << "Ts #" << i << endl;
//		int j = 0;
//		for (mm_Iter = mainMem->MM.begin(); mm_Iter != mainMem->MM.end(); j++) {
//			if (offset + i == *indxPtrIter) { //the main memory element is the current DB element that we check. Skip it.
//				mm_Iter++;
//				indxPtrIter++;
//				nndistIter++;
//				nnbsIter++;
//				continue;
//			}
//			double dist = nnDist(dbEl, *mm_Iter, true, *nndistIter);
//			if (dist < sqrThreshold) {
//				float* mmEl = *mm_Iter;
//				delete [] mmEl;
//				mainMem->MM.erase(mm_Iter);
//				mainMem->indxPtr.erase(indxPtrIter);
//				nn_dists.erase(nndistIter);
//				nnbs.erase(nnbsIter);
//				tmp_avr_comparisons += (j+1);
//			} else {
//				if ( dist < *nndistIter) {
//					*nndistIter = (float) dist;
//					*nnbsIter = offset+i;
//				}
//				mm_Iter++;
//				indxPtrIter++;
//				nndistIter++;
//				nnbsIter++;
//			}
//		}
//	}
//	tmp_avr_comparisons /= (int) DB.size();
//	avr_comparisons.push_back(tmp_avr_comparisons);
//}

void DiscDiscords::updateNNbs(vector<float*> &DB, int offset) {
	double sqrThreshold = threshold*threshold;
	int tmp_avr_comparisons = 0; //how many comparisons on average are necessary before discarding a DB element
	for (int i = 0; i < (int) DB.size(); i++) {
		float* dbEl = DB[i];
		int j = 0;
		for (int p = 0; p < (int) mainMem->MM.size(); p++, j++) {
			if (offset + i == mainMem->indxPtr[p]) { //the main memory element is the current DB element that we check. Skip it.
				continue;
			}
			double dist = nnDist(dbEl, mainMem->MM[p], true, nn_dists[p]);
			if (dist < sqrThreshold) {
				float* mmEl = mainMem->MM[p];
				delete [] mmEl;
				mainMem->MM.erase(mainMem->MM.begin() + p);
				mainMem->indxPtr.erase(mainMem->indxPtr.begin() + p);
				nn_dists.erase(nn_dists.begin() + p);
				nnbs.erase(nnbs.begin() + p);
				tmp_avr_comparisons += (j+1);
				p--;
			} else {
				if (dist < nn_dists[p]) {
					nn_dists[p] = (float) dist;
					nnbs[p] = offset+i;
				}
			}
		}
	}
	tmp_avr_comparisons /= (int) DB.size();
	avr_comparisons.push_back(tmp_avr_comparisons);
}

void DiscDiscords::selectTopDiscords() {
	vector<int> sortedNbs;
	for (int i = 0; i < k_disc; i++) {
		if (mainMem->MM.size() == 0) {
			//the memory was emptied before finding the required number of discords
			break;
		}
		float maxDist = 0;
		int maxPos = 0;
		for (int j = 0; j < (int) mainMem->MM.size(); j++) {
			float tmpDist = nn_dists[j];
			if (tmpDist > maxDist) {
				maxDist = tmpDist;
				maxPos = j;
			}
		}
		float* discord = mainMem->MM[maxPos];
		discords.push_back(discord);
		int ptr = mainMem->indxPtr[maxPos];
		discord_indxPtrs.push_back(ptr);
		float dst = nn_dists[maxPos];
		discord_nndists.push_back(sqrt(dst));
		sortedNbs.push_back(nnbs[maxPos]);

		mainMem->MM.erase(mainMem->MM.begin() + maxPos);
		mainMem->indxPtr.erase(mainMem->indxPtr.begin() + maxPos);
		nn_dists.erase(nn_dists.begin() + maxPos);
		nnbs.erase(nnbs.begin() + maxPos);
	}
	nnbs.clear();
	for (int i = 0; i < (int) sortedNbs.size(); i++) {
		nnbs.push_back(sortedNbs[i]);
	}
}

void DiscDiscords::loadDB(vector<float*>& DB, int pageNum) {
	// generate the file name
	string fileName(dir);
	fileName += "TS";
	stringstream oss;
    oss << fileName << pageNum;
	fileName = oss.str();
	fileName += ".txt";

	read_matrix1(DB, fileName, page_size, ts_len);
}

double DiscDiscords::nnDist(float* x, float* y, bool earlyAbandon, double minDist) {
	double* dist_numoper = NULL;
	switch (dist_type) {
		case Globals::EUCLID_DIST : {
			dist_numoper = sqreuclid_distf(x, y, ts_len, earlyAbandon, minDist);
			break;
		}
		case Globals::ROTAT_DIST: {
			dist_numoper = sqr_rotatdist(x, y, ts_len, earlyAbandon, minDist);
			break;
		}
	}
	//todo: store the number of performed operations
	double dist = dist_numoper[0];
	delete [] dist_numoper;
	return dist;
}

void DiscDiscords::logDiscords() {
	for (int i = 0; i < k_disc; i++) {
		if (i < 10) { // cout just the first 10 discords
			cout << "------------Discord# " << i << "----------" << endl;
			cout << "Id:" << discord_indxPtrs[i] << endl;
			cout << "NNdist=" << discord_nndists[i] << endl;
			cout << "NNbr = " << nnbs[i] << endl;
			cout << "-------------------------------------" << endl;
		}
		DISCINDSLOG << discord_indxPtrs[i] << "\t" << discord_nndists[i] << "\t" << nnbs[i] << endl;
		float* discord = discords[i];
		for (int j = 0; j < this->ts_len; j++) {
			DISCORDSLOG << "" << discord[j] << " ";
		}
		DISCORDSLOG << endl;
	}
}
