#include "main_memory.h"
#include "disc_discords.h"
#include "globals.h"
#include "utils.h"
#include "io_utils.h"
#include <iostream>
#include <sstream>
#include <ctime>
#include <complex>

using namespace std;

MainMemory::MainMemory(const char* dirName, double threshold, int mem_pages, int dist_type, int page_size, int ts_len) {
	this->dir = dirName;
	this->threshold = threshold;
	this->mem_pages = mem_pages;
	this->ts_len = ts_len;
	this->page_size = page_size;
	this->dist_type = dist_type;
	this->usePivots = Globals::USE_PIVOTS;
	loadPivots();
}

MainMemory::~MainMemory() {
	MMvolume.empty();
	indxPtr.empty();
	cleanContainer(MM);
	cleanContainer(pivots);
	cleanContainer(pivotDists);
}

void MainMemory::buildMemory() {
	int starttimer = clock();
	cout << "In build memory...." << endl;
    int offset = 0;
	vector <float*> DB;
	for (int i = 0; i < mem_pages; i++){
        cout << "---Building MainMemory: page#" << (i+1) << " of " << mem_pages;
		if (mem_pages > 1) {
			this->page_size = -1;
			this->ts_len = -1;
		}
        loadDB(DB, i);
		cout << " .... len=" << page_size << "; width=" << ts_len << endl;
		//cout << "DB.size = " << (int) DB.size() << endl;
        buildMainMemory(DB, offset);
		(*MEMLOG) << MMvolume[i] << " " << avr_comparisons[i] << endl;
        offset += (int) DB.size();
		cleanContainer(DB);
	}
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	cout << "(BuildMemory)Time spent: " << time_spent << endl;
	cout << "End memmory size = " << MMvolume[MMvolume.size()-1] << endl;
	(*LOG) << "(BuildMemory)Time spent: " << time_spent << "sec." << endl;
	(*LOG) << "End memmory size = " << MMvolume[MMvolume.size()-1] << endl;
	//dumpMemory();
}

void MainMemory::buildMainMemory(vector<float*>& DB, int newoffset) {
    int offset = newoffset;
	double sqrThreshold = threshold*threshold;
	int tmp_avr_comparisons = 0;
	for (int i = 0; i < (int) DB.size(); i++) {
		//if (! (i % 100)) {
		//	cout << "DB element" << i << endl;
		//}
		float* dbEl = DB[i];
		bool failsThreshold = 0;
		float* el_pivotDists = NULL;
		if (Globals::USE_PIVOTS) {
			el_pivotDists = getPivotDists(dbEl, failsThreshold);
			if (failsThreshold) {
				deletePivotDists(el_pivotDists);
				continue;
			}
		}
		bool addEl = true;
		vector <float*>::iterator mm_Iter = MM.begin();
		vector <int>::iterator indxPtrIter = indxPtr.begin();
		//cout << "Ts #" << i << endl;
		int pos = 0;
		for (mm_Iter = MM.begin(); mm_Iter != MM.end(); ) {
			bool exceeds = false;
			if (Globals::USE_PIVOTS) {
				exceeds = exceedsPivot(el_pivotDists, pivotDists[pos], threshold);
			}
			if (!exceeds) { //we have to perform the NN distance computation
				tmp_avr_comparisons++;
			}
			if (!exceeds && nnDist(dbEl, *mm_Iter, true, sqrThreshold) < sqrThreshold) {//remove the element, can't be a discord
				float* mmEl = *mm_Iter;
				delete [] mmEl;
				MM.erase(mm_Iter);
				indxPtr.erase(indxPtrIter);
				deletePivotDists(pos);
				addEl = false;
				if (!Globals::CHECKALL_MODE) {
					break;
				}
				//else {
				//	cout << "globals.checkall = " << Globals::CHECKALL_MODE << endl;
				//}
			} else {
				mm_Iter++;
				indxPtrIter++;
				pos++;
			}
		}
		//if (addEl && farFromPivots(el_pivotDists, threshold)) { no need, we've already checked that it is far from the pivots
		if (addEl) {
			float* dbEl_cpy = new float[ts_len];
			for (int j = 0; j < ts_len; j++) {
				dbEl_cpy[j] = dbEl[j];
			}
			MM.push_back(dbEl_cpy);
			indxPtr.push_back(offset + i);
			addPivotDists(el_pivotDists);
		} else {
			deletePivotDists(el_pivotDists);
		}
	}
	tmp_avr_comparisons /= (int) DB.size();
	avr_comparisons.push_back(tmp_avr_comparisons);
    MMvolume.push_back((int) MM.size());
}

bool MainMemory::farFromPivots(float* el_dists, double threshold) {
	if (Globals::USE_PIVOTS) {
		for (int i = 0; i < (int) pivots.size(); i++) {
			if (el_dists[i] < threshold && el_dists[i] != 0) {
				return false;
			}
		}
	}
	return true;
}

bool MainMemory::exceedsPivot(float* el_dists, float* dists, double threshold) {
	if (Globals::USE_PIVOTS && el_dists != NULL) {
		int pivotNum = (int) pivots.size();
		for (int i = 0; i < pivotNum; i++) {
			if (abs(dists[i] - el_dists[i]) > threshold) {
				return true;
			}
		}
	}
	return false;
}

void MainMemory::addPivotDists(float* el_dists) {
	if (usePivots) {
		pivotDists.push_back(el_dists);
	}
}
    
void MainMemory::deletePivotDists(float* el_pivotDists) {
	if (usePivots) {
		delete [] el_pivotDists;
	}
}

void MainMemory::deletePivotDists(int memElId) {
	if (usePivots) {
		float* dists = pivotDists[memElId];
		pivotDists.erase(pivotDists.begin() + memElId);
		delete [] dists;
	}
}

float* MainMemory::getPivotDists(float* dbEl, bool& failsThreshold) {
	if (!usePivots)  {
		return NULL;
	}
	float* dists = new float[pivots.size()];
	for (int i = 0; i < (int) pivots.size(); i++) {
		dists[i] = (float) sqrt(this->nnDist(dbEl, pivots[i], false, 0));			
		if (dists[i] != 0 && dists[i] < this->threshold) {
			failsThreshold = true;
			break;
		}
	}
	return dists;
}

void MainMemory::loadPivots() {
	if (usePivots) {
		string fileName(dir);
		fileName += "Output_pivots.txt";
		int numPivots = -1;
		int pivotLen = -1;
		read_matrix1(pivots, fileName, numPivots, pivotLen);
	}
}

void MainMemory::loadDB(vector<float*>& DB, int pageNum) {
	// generate the file name
	string fileName(dir);
	fileName += "TS";
	stringstream oss;
    oss << fileName << pageNum;
	fileName = oss.str();
	fileName += ".txt";

	read_matrix1(DB, fileName, page_size, ts_len);
}

double MainMemory::nnDist(float* x, float* y, bool earlyAbandon, double minDist) {
	double* dist_numoper = NULL;
	switch (dist_type) {
		case Globals::EUCLID_DIST : {
			dist_numoper = sqreuclid_distf(x, y, ts_len, earlyAbandon, minDist);
			break;
		}
		case Globals::ROTAT_DIST: {
			dist_numoper = sqr_rotatdist(x, y, ts_len, earlyAbandon, minDist);
			break;
		}
	}
	//todo: store the number of performed operations
	double dist = dist_numoper[0];
	delete [] dist_numoper;
	return dist;
}

void MainMemory::dumpMemory() {
	string strdir = dir;
    MEMDUMP1.open((strdir + "Output_memDump1.txt").c_str());
	MEMDUMP2.open((strdir + "Output_memDump2.txt").c_str());
	for (int i = 0; i < (int) MM.size(); i++) {
		MEMDUMP2 << indxPtr[i] << endl;
		float* discord = MM[i];
		for (int j = 0; j < this->ts_len; j++) {
			MEMDUMP1 << "" << discord[j] << " ";
		}
		MEMDUMP1 << endl;
	}
	MEMDUMP1.close();
	MEMDUMP2.close();
}
