// cpp_discords.cpp : Defines the entry point for the console application.
//

#include "disc_discords.h"
#include "threshold_estimator.h"
#include "NNEstimator.h"
#include "globals.h"
#include "test.h"
#include <iostream>
#include <fstream>

using namespace std;

int type = 1;
int db_pages = 1;
double threshold = 1;
double sampleSize = 0;
double db_size = 0;
double num_discords = 0;

const char* folder;
const char* filename;


void printUsage() {
	cout << "cpp_discords -t <type> [options]" << endl;
	cout << "\t type:" << endl;
	cout << "\t      - 0 compute threshold" << endl;
	cout << "\t      - 1 find discords" << endl;
	cout << "\t      - 2 find the nearest neighbor to a set of timeseries. The indices of the time series are stored as column in <folder>/Output_indsNND" << endl;
	cout << "\t options:" << endl; 
	cout << "\t -r   - threshold (used only with -t 1)" << endl;
	cout << "\t -d   - number of discords to be detected on the sample" << endl;
	cout << "\t -f   - folder for input/output files" << endl;
	cout << "\t -p   - number of db pages(used only with -t 1)" << endl;
	cout << "\t -n   - data base size" << endl;
	//cout << "\t -s   - input file name (used only with -d 1)" << endl;
	cout << "\t -z   - sample size for computing the threshold" << endl;
	cout << endl;
}

void parseArgs(const int argc, char* argv[]) {
	for (int i = 1; i < argc; i+=2) {
		if (!strcmp("-t", argv[i])) {
			type = atoi(argv[i+1]);
			continue;
		} 
		if (!strcmp("-r", argv[i])) {
			threshold = atof(argv[i+1]);
			continue;
		}
		if (!strcmp("-d", argv[i])) {
			num_discords = atof(argv[i+1]);
			continue;
		}
		if (!strcmp("-f", argv[i])) {
			folder = argv[i+1];
			continue;
		}
		if (!strcmp("-p", argv[i])) {
			db_pages = atol(argv[i+1]);
			continue;
		}
		if (!strcmp("-n", argv[i])) {
			db_size = atof(argv[i+1]);
			continue;
		}
		if (!strcmp("-s", argv[i])) {
			filename = argv[i+1];
			continue;
		}
		if (!strcmp("-z", argv[i])) {
			sampleSize = atof(argv[i+1]);
			continue;
		}
	}
}

void runTest() {
	Test* t = new Test();
	t->run();
	delete t;
}

void readTao() {
	const int numberOfStreams = 12218;
	const int windowSize = 962;
	long int currentAddresses[numberOfStreams];
	double firstWindow[windowSize];
	FILE* fp;
	if((fp=fopen("../data/heterogeneous/long_time_series/tao.dat","r+b"))==NULL ) {
		printf("\n File error");
	}

	//read the header
	for(int i=0;i<numberOfStreams;i++){
		fread(&i,sizeof(int),1,fp);
		fread(&currentAddresses[i],sizeof(long int),1,fp);
	}

	FILE* outFile;
	outFile = fopen("../data/heterogeneous/long_time_series/tao_all.dat", "w");
	//read the first "windowsize" values of stream i
	for(int i=0;i<numberOfStreams;i++){
		cout << "Storing stream " << i << " of " << numberOfStreams << endl;
		fseek(fp,currentAddresses[i],SEEK_SET);
		fread(firstWindow,sizeof(double),windowSize,fp);
		for (int j = 0; j < windowSize; j++) {
			fprintf(outFile, "%8.4f\n", firstWindow[j]);
		}
	}
	fclose(fp);
	fclose(outFile);
}

int main(int argc, char* argv[]) {
	//readTao();
	//runTest();
	//return 0;
	if (argc < 2 || !(argc % 2)) {
		printUsage();
		return(0);
	} 
	parseArgs(argc, argv);

	if (type == 0) {
    //-t 0 -r 10 -f ../data/randomwalk1/ -p 10 -n 10000 -z 5000
	// call the threshold estimation tool
		int dist_type = Globals::EUCLID_DIST;
        ThresholdEstimator* testim = 
			new ThresholdEstimator(folder, filename, db_size, sampleSize, 
			  					   db_pages, dist_type, num_discords);
		testim->computeThreshold1();
	//	testim->computeThreshold1();
		double threshold = testim->threshold;
		cout << "Threshold = " << threshold << endl;
		delete testim;
	} else if (type == 1) {
	// call the discord detection tool
	//-t 1 -r 10 -d 14.5067 -f ../data/randomwalk1/ -p 10
		cout << "Threshold = " << threshold << endl;
		//DiscDiscords* dd = new DiscDiscords(dir, top_discords, threshold, db_pages, DiscDiscords::EUCLID_DIST, page_size, ts_length);
		DiscDiscords* dd = new DiscDiscords(folder, (int) num_discords, threshold, db_pages, Globals::EUCLID_DIST);
		dd->findDiscords();
		delete dd;
	} else {
	// find the nearest neighbors to a set of time series
	//-t 2 -f ../data/randomwalk1/ -p 10
		NNEstimator *nne = new NNEstimator(folder, db_pages, Globals::EUCLID_DIST);
		nne->computeNNDists();
		delete nne;
	}
	return 0;
}

