#ifndef __THRESHOLDESTIMATOR_H
#define __THRESHOLDESTIMATOR_H

#include <vector>
#include <fstream>

using namespace std;
class ThresholdEstimator {
public:
	float threshold;
	const char* dir;
	const char* file;
	double sampleSize;
	double db_size;
	int db_pages;


	double num_discords;
	double epsilon;   //slack required by Chernoff's bound. The algorithm guarantees that 
		              //with high probability there will be between num_discords and numdiscords + 2*epsilon 
					  //elements when the true distribution is considered

	vector<float*> sample;
	vector<float*> sampleDb;
	vector<float*> pivots;
	vector<float*> nnSampleDists;

private:
	int position;
	int lineCounter;
	int ts_size;
	int ts_sizeconstant;
	int page_size;
	int dist_type;
	ofstream LOG;
	ofstream LOG1;
	ofstream PIVOTLOG;
	bool usePivots;

public:
	ThresholdEstimator(const char* dir, const char* file, double db_size, 
					   double sampleSize, int db_pages, int dist_type, double num_discords);
	~ThresholdEstimator();

	void computeThreshold();
	void computeThreshold1();

private:
	void addToSampleInds(int inds);
	void addToSampleInds(int inds, vector<float*>& smpl);
	void setTsSize(char* line);
	float* readExample(char* line);
	void loadSample(int pageNum);
	void buildSample();
	void computeNNDists();
	void loadDB(vector<float*>& DB, int pageNum);
	void updateNNbs(vector<float*> &DB, int offset);
	double nnDist(float* x, float* y, bool earlyAbandon, double minDist);
	void computeRndInds();
	void computeSampleNNDists();
	void computeThresholdFromDists();
	void computeEpsilon();
	void logPivots();
	void collectGoodPivots();
	void collectGoodPivots1(int numPivots);
	void sortSamplesDists();
	float* minAvrDistPivot();
	float* furthestFromPivot(int pivotInd);
};

#endif