#include <math.h>
#include "utils.h"

using namespace std;

double round(double x) {
	double diff = x - floor(x);
	if (diff >= 0.5){
		return ceil(x);
	}else{
		return floor(x);
	}
}

double mean(double* data, int start, int end) {
    double sum = 0;
    for (int i = start; i < end; i++) {
        sum += data[i];
    } 
    return sum / (end - start);
}

double stdev(double* data, int start, int end, double m, bool hasMean) {
    int n = end - start;
    if (n == 1) {
        return data[start];
    }

    if (!hasMean) {
        m = mean(data, start, end);
    }
    double sum = 0;
    double nextel = 0;
    for (int i = start; i < end; i++) {
        nextel = (data[i] - m);
        sum += nextel * nextel;
    }
    sum /= (n - 1);
    return sqrt(sum);
}

double stdev(double* data, int start, int end) {
    return stdev(data, start, end, 0, false);
}

void znormalize(double* data, int start, int end) {
    int n = end - start;
    double m = mean(data, start, end);
    double s = stdev(data, start, end, m, true);
    for (int i = start; i < end; i++) {
        data[i] = (data[i] - m) / s;
    }
}

void arraycopy(double* src, int srcpos, double* dst, int dstpos, int size) {
    for (int i = 0; i < size; i++) {
        dst[dstpos+i] = src[srcpos+i];
    }
}

double* sqreuclid_dist(double* x, double* y, int size, bool earlyAbandon, double mindist) {
    double dist = 0;
    double operations = 0;
    for (int i = 0; i < size; i++) {
        if (earlyAbandon && mindist < dist) {
            break;
        }
        operations++;
        dist += (x[i]-y[i])*(x[i]-y[i]);
    }
    double* res = new double[2];
    res[0] = dist; res[1] = operations;
    return res;
}

double* sqreuclid_distf(float* x, float* y, int size, bool earlyAbandon, double mindist) {
    double dist = 0;
    double operations = 0;
    for (int i = 0; i < size; i++) {
        if (earlyAbandon && mindist < dist) {
            break;
        }
        operations++;
        dist += (x[i]-y[i])*(x[i]-y[i]);
    }
    double* res = new double[2];
    res[0] = dist; res[1] = operations;
    return res;
}

double* sqr_rotatdist(float* x, float* y, int size, bool earlyAbandon, double mindist) {
	double rdist = mindist;
	double operations = 0;
	for (int i = 0; i < size; i++) {
		double* tmp_dist_oper = shifted_sqreucliddist(x, y, i, size, earlyAbandon, rdist);
		if (tmp_dist_oper[0] < rdist) {
			rdist = tmp_dist_oper[0];
		}
		operations += tmp_dist_oper[1];
		delete [] tmp_dist_oper;
	}
	double* res = new double[2];
	res[0] = rdist; res[1] = operations;
	return res;
}

double* shifted_sqreucliddist(float* x, float* y, int posy, int size, bool earlyAbandon, double mindist) {
	double dist = 0;
	double operations = 0;
    for (int i = 0; i < size; i++) {
        if (earlyAbandon && mindist < dist) {
            break;
        }
        operations++;
		if (posy == size) {
			posy = 0;
		}
        dist += (x[i]-y[posy])*(x[i]-y[posy]);
		posy++;
    }
    double* res = new double[2];
    res[0] = dist; res[1] = operations;
    return res;
}

void cleanContainer(vector <float*>& cont) {
	while (cont.size() > 0) {
		float* vEl = cont[0];
		cont.erase(cont.begin());
		delete [] vEl;
	}
}